﻿;xinput.pb
;(c) 2014 Roger Hågensen
;zlib License, http://en.wikipedia.org/wiki/Zlib_License

;This code is what you would use if the user selects "Xinput" controller mapping in your software.

;References
;http://msdn.microsoft.com/en-us/library/windows/desktop/hh405053.aspx
DeclareModule XInput
  #XINPUT_GAMEPAD_DPAD_UP            = $0001
  #XINPUT_GAMEPAD_DPAD_DOWN         = $0002
  #XINPUT_GAMEPAD_DPAD_LEFT         = $0004
  #XINPUT_GAMEPAD_DPAD_RIGHT         = $0008
  #XINPUT_GAMEPAD_START            = $0010
  #XINPUT_GAMEPAD_BACK               = $0020
  #XINPUT_GAMEPAD_LEFT_THUMB         = $0040
  #XINPUT_GAMEPAD_RIGHT_THUMB      = $0080
  #XINPUT_GAMEPAD_LEFT_SHOULDER      = $0100
  #XINPUT_GAMEPAD_RIGHT_SHOULDER   = $0200
  #XINPUT_GAMEPAD_A                  = $1000
  #XINPUT_GAMEPAD_B                  = $2000
  #XINPUT_GAMEPAD_X                  = $4000
  #XINPUT_GAMEPAD_Y                  = $8000
  Structure XINPUT_NORMALIZED
   buttons.l   ;See #XINPUT_GAMEPAD_ masks.
   triggerl.f   ;0.0 to 1.0
   triggerr.f   ;0.0 to 1.0
   thumblx.f   ;Left thumbstick x-axis value, -1.0 to +1.0, 0.0 is centered.
   thumbly.f   ;Left thumbstick y-axis value, -1.0 to +1.0, 0.0 is centered.
   thumbrx.f   ;Right thumbstick x-axis value, -1.0 to +1.0, 0.0 is centered.
   thumbry.f   ;Right thumbstick y-axis value, -1.0 to +1.0, 0.0 is centered.
  EndStructure
  Macro Xinput_Button(buttons, mask)
     Bool(buttons & mask)
  EndMacro
  Declare Xinput_Init()
  Declare Xinput_Get(xinput, index.l)
  Declare Xinput_Free(xinput)
  Declare Xinput_Version(xinput)
  Declare Xinput_Vibrate(xinput, index.l, low.f, high.f)
EndDeclareModule

Module XInput
  EnableExplicit
  
  Prototype XInputGetState(dwUserIndex.l, *pState)
  Prototype XInputSetState(dwUserIndex.l, *pVibration)
  
               
  
  Structure XINPUT_GAMEPAD
     wButtons.w          ;See #XINPUT_GAMEPAD_ masks.
     bLeftTrigger.a      ;A value of 0 to 255.
     bRightTrigger.a   ;A value of 0 to 255.
     sThumbLX.w         ;Left thumbstick x-axis value, -32768 to +32767, 0 is centered.
     sThumbLY.w         ;Left thumbstick y-axis value, -32768 to +32767, 0 is centered.
     sThumbRX.w         ;Right thumbstick x-axis value, -32768 to +32767, 0 is centered.
     sThumbRY.w         ;Right thumbstick y-axis value, -32768 to +32767, 0 is centered.
  EndStructure
  
  Structure XINPUT_STATE
     dwPacketNumber.l
     Gamepad.XINPUT_GAMEPAD
  EndStructure
  
  Structure XINPUT_VIBRATION
     wLeftMotorSpeed.w ;Low frequency rumble, 0 to 65535, 0 is off and 65535 is 100% speed.
     wRightMotorSpeed.w ;High frequency rumble, 0 to 65535, 0 is off and 65535 is 100% speed.
  EndStructure
  
  Structure XINPUT_XINPUT
     dll.i
     version.i
     *getstate.XInputGetState
     *setstate.XInputSetState
     state.XINPUT_STATE[4]
     normalized.XINPUT_NORMALIZED[4]
  EndStructure
  

  
  Procedure.i Xinput_Init()
     Protected *xinput.XINPUT_XINPUT
  
     *xinput = AllocateMemory(SizeOf(XINPUT_XINPUT))
     If *xinput = #False
        ProcedureReturn #Null
     EndIf
  
     *xinput\dll = OpenLibrary(#PB_Any, "xinput1_4.dll") ;Windows 8+
  
     *xinput\version = 14
     If *xinput\dll = #False
        *xinput\dll = OpenLibrary(#PB_Any, "xinput1_3.dll") ;DirectX SDK
        *xinput\version = 13
     EndIf
     If *xinput\dll = #False
        *xinput\dll = OpenLibrary(#PB_Any, "xinput1_2.dll") ;DirectX SDK
        *xinput\version = 12
     EndIf
     If *xinput\dll = #False
        *xinput\dll = OpenLibrary(#PB_Any, "xinput1_1.dll") ;DirectX SDK
        *xinput\version = 11
     EndIf
     If *xinput\dll = #False
        *xinput\dll = OpenLibrary(#PB_Any, "xinput9_1_0.dll") ;Windows Vista/7/8
        *xinput\version = 10
     EndIf
     If *xinput\dll = #False
        *xinput\version = 0
     EndIf
  
     *xinput\getstate = GetFunction(*xinput\dll, "XInputGetState")
     If *xinput\getstate = #Null
        *xinput\version = 0
     EndIf
  
     *xinput\setstate = GetFunction(*xinput\dll, "XInputSetState")
     If *xinput\setstate = #Null
        *xinput\version = 0
     EndIf
     
     If (*xinput\version = 0) And (*xinput\dll <> #Null)
        CloseLibrary(*xinput\dll)
     EndIf
     If *xinput\version = 0
        FreeMemory(*xinput)
        *xinput = #Null
     EndIf
  
     ProcedureReturn *xinput
  EndProcedure
  
  Procedure.i Xinput_Free(*xinput.XINPUT_XINPUT)
    If *xinput = #NUL : ProcedureReturn : EndIf
     If *xinput = #Null
        ProcedureReturn #Null
     EndIf
     
     If *xinput\dll <> #Null
        If IsLibrary(*xinput\dll)
           CloseLibrary(*xinput\dll)
        EndIf
        *xinput\dll = #Null
     EndIf
     
     FreeMemory(*xinput)
  
     ProcedureReturn #Null
  EndProcedure
  
  Procedure.i Xinput_Version(*xinput.XINPUT_XINPUT)
    If *xinput = #NUL : ProcedureReturn #False : EndIf
     ProcedureReturn *xinput\version
  EndProcedure
  
  #ERROR_DEVICE_NOT_CONNECTED = 1167
  
  Procedure.i Xinput_Get(*xinput.XINPUT_XINPUT, index.l)
    If *xinput = #NUL : ProcedureReturn : EndIf
     Protected error.i
     If (index < 0) Or (index > 3)
        ProcedureReturn #Null
     EndIf
     ZeroMemory_(*xinput\state[index], SizeOf(XINPUT_STATE))
     ZeroMemory_(*xinput\normalized[index], SizeOf(XINPUT_NORMALIZED))
     error = *xinput\getstate(index, *xinput\state[index])
     If error <> #ERROR_SUCCESS
        ProcedureReturn #Null
     EndIf
     *xinput\normalized[index]\buttons = *xinput\state[index]\Gamepad\wButtons
     *xinput\normalized[index]\triggerl = *xinput\state[index]\Gamepad\bLeftTrigger / 255.0
     *xinput\normalized[index]\triggerr = *xinput\state[index]\Gamepad\bRightTrigger / 255.0
     If *xinput\state[index]\Gamepad\sThumbLX < 0
        *xinput\normalized[index]\thumblx = *xinput\state[index]\Gamepad\sThumbLX / 32768.0
     Else
        *xinput\normalized[index]\thumblx = *xinput\state[index]\Gamepad\sThumbLX / 32767.0
     EndIf
     If *xinput\state[index]\Gamepad\sThumbLY < 0
        *xinput\normalized[index]\thumbly = *xinput\state[index]\Gamepad\sThumbLY / 32768.0
     Else
        *xinput\normalized[index]\thumbly = *xinput\state[index]\Gamepad\sThumbLY / 32767.0
     EndIf
     If *xinput\state[index]\Gamepad\sThumbRX < 0
        *xinput\normalized[index]\thumbrx = *xinput\state[index]\Gamepad\sThumbRX / 32768.0
     Else
        *xinput\normalized[index]\thumbrx = *xinput\state[index]\Gamepad\sThumbRX / 32767.0
     EndIf
     If *xinput\state[index]\Gamepad\sThumbRY < 0
        *xinput\normalized[index]\thumbry = *xinput\state[index]\Gamepad\sThumbRY / 32768.0
     Else
        *xinput\normalized[index]\thumbry = *xinput\state[index]\Gamepad\sThumbRY / 32767.0
     EndIf
     ProcedureReturn *xinput\normalized[index]
  EndProcedure
  
  Procedure.i Xinput_Vibrate(*xinput.XINPUT_XINPUT, index.l, low.f, high.f)
    If *xinput = #NUL : ProcedureReturn : EndIf
     Protected error.i, motor.XINPUT_VIBRATION
     If (index < 0) Or (index > 3)
        ProcedureReturn #Null
     EndIf
     motor\wLeftMotorSpeed = low * 65535.0
     motor\wRightMotorSpeed = high * 65535.0
     error = *xinput\setstate(index, motor)
     If error <> #ERROR_SUCCESS
        ProcedureReturn #Null
     EndIf
     ProcedureReturn #ERROR_SUCCESS
  EndProcedure

EndModule
;************************************************************************
;Example
CompilerIf #PB_Compiler_IsMainFile
  Define *xinput, *state.XInput::XINPUT_NORMALIZED
  
  If OpenConsole("Xinput Test") = #False
     End
  EndIf
  EnableGraphicalConsole(#True)
  
  *xinput = XInput::Xinput_Init()
  If *xinput = #Null
     PrintN("Xinput_Init() failed.")
     End
  EndIf
  
  ;Note that with DirectInput the Left and Right triggers are considered being on the same (Z) axis, with Xinput they are not.
  
  ; If a Xinput gamepad is connected,
  ; the expected results with Xinput_Get() are...
  ; Button A: 0 or 1
  ; Button B: 0 or 1
  ; Button X: 0 or 1
  ; Button Y: 0 or 1
  ; Button Left Shoulder: 0 or 1
  ; Button Right Shoulder: 0 or 1
  ; Button Back: 0 or 1
  ; Button Start: 0 or 1
  ; Button Left Thumb: 0 or 1
  ; Button Right Thumb: 0 or 1
  ; Button D-Pad Left: 0 or 1
  ; Button D-Pad Up: 0 or 1
  ; Button D-Pad Right: 0 or 1
  ; Button D-Pad Down: 0 or 1
  ; Left Trigger: 0.0 to 1.0
  ; Right Trigger: 0.0 to 1.0
  ; Left Stick X: -1.0 (left) to 0.0 (center) to 1.0 (right)
  ; Left Stick Y: -1.0 (down) to 0.0 (center) to 1.0 (up)
  ; Right Stick X: -1.0 (left) to 0.0 (center) to 1.0 (right)
  ; Right Stick Y: -1.0 (down) to 0.0 (center) to 1.0 (up)
  
  Define timeold.l, timenew.l, vibrate.i
  
  timenew = ElapsedMilliseconds()
  timeold = timenew
  
  Repeat
     timenew = ElapsedMilliseconds()
     *state = XInput::Xinput_Get(*xinput, 0)
     If *state
        ClearConsole()
        PrintN("Button A: " + Str(XInput::Xinput_Button(*state\buttons, XInput::#XINPUT_GAMEPAD_A)))
;         PrintN("Button B: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_B)))
;         PrintN("Button X: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_X)))
;         PrintN("Button Y: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_Y)))
;         PrintN("Button Left Shoulder: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_LEFT_SHOULDER)))
;         PrintN("Button Right Shoulder: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_RIGHT_SHOULDER)))
;         PrintN("Button Back: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_BACK)))
;         PrintN("Button Start: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_START)))
;         PrintN("Button D-Pad Left: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_DPAD_LEFT)))
;         PrintN("Button D-Pad Up: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_DPAD_UP)))
;         PrintN("Button D-Pad Right: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_DPAD_RIGHT)))
;         PrintN("Button D-Pad Down: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_DPAD_DOWN)))
;         PrintN("Button Left Thumb: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_LEFT_THUMB)))
;         PrintN("Button Right Thumb: " + Str(Xinput_Button(*state\buttons, #XINPUT_GAMEPAD_RIGHT_THUMB)))
        PrintN("Left Trigger: " + StrF(*state\triggerl, 7) )
        PrintN("Right Trigger: " + StrF(*state\triggerr, 7))
        PrintN("Left Stick X: " + StrF(*state\thumblx, 7))
        PrintN("Left Stick Y: " + StrF(*state\thumbly, 7))
        PrintN("Right Stick X: " + StrF(*state\thumbrx, 7))
        PrintN("Right Stick Y: " + StrF(*state\thumbry, 7))
        PrintN("")
        PrintN("Press enter to stop.")
     Else
        PrintN("Error")
        Break
     EndIf
     If (timenew - timeold) >= 2000
        timeold = timenew
        If vibrate = 1
           XInput::Xinput_Vibrate(*xinput, 0, 0.5, 0.4)
           vibrate = 0
        Else
           XInput::Xinput_Vibrate(*xinput, 0, 0.0, 0.0)
           vibrate = 1
        EndIf
     EndIf
     Delay(100)
  Until Inkey() <> ""
  XInput::Xinput_Vibrate(*xinput, 0, 0.0, 0.0)
  
  PrintN("")
  PrintN("Xinput DLL Version: " + StrD(XInput::Xinput_Version(*xinput) * 0.1, 1))
  PrintN("Press enter to quit.")
  Input()
  
  *xinput = XInput::Xinput_Free(*xinput)
  CloseConsole()
  End
CompilerEndIf
; IDE Options = PureBasic 5.70 LTS (Windows - x64)
; CursorPosition = 197
; FirstLine = 69
; Folding = H9
; EnableXP